#include "CCommon.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <algorithm>

// ------------------------------------------------------------------
// CSPort
// ------------------------------------------------------------------
CSPort::CSPort()
{
  // Configure interrupts
  this->IntNo        = -1;
  this->IntDataCount = -1;

  // Type (default == unused) & default number
  this->Type   = PORT_UNUSED;
  this->PortNo = -1;
  _snprintf(this->PortName, sizeof(this->PortName), "PortBus(unnammed)");

  // Configure state
  this->StateData = 0x00;

  // Configure onwer
  this->OwnerHw      = NULL;

  // Configure buffer
  this->Buffer.resize(PORT_BUFFER_SIZE);
  memset(&this->Buffer, 0, PORT_BUFFER_SIZE);
  this->BufferDataCount = 0;
}

const char* 
CSPort::GetName() const
{
  // This is the port bus ;)
  return this->PortName;
}


CHardware*
CSPort::GetOwner() const
{
  // Return owner
  return this->OwnerHw;
}

void
CSPort::SetPortNo(int PortNo)
{
  // Just set port number
  this->PortNo = PortNo;

  // And cache the port name
  _snprintf(this->PortName, sizeof(this->PortName), "PortBus($%.4x)",
      (unsigned short)PortNo);
}


bool
CSPort::IsDataPresent() const
{
  // What kind of a port are we dealing with?
  switch(this->Type)
  {
    // Buffered? Then check the BufferDataCount
    case PORT_BUFFERED: return (bool)(BufferDataCount > 0);

    // State? Then there is always sth to read :)
    case PORT_STATE: return true;

    // Unused / unsupported?
    default:
      break;
  }

  // No data I guess
  return false;
}

bool
CSPort::Recv(uint8_t *RetByte)
{
  // Dispatch
  switch(this->Type)
  {
    case PORT_BUFFERED: return this->BufferedRecv(RetByte);
    case PORT_STATE:    return this->StateRecv(RetByte);
    default: break;
  }

  // Seems unused or unsupported
  CLOG(LOG_WARNING, "Tried to receive data from an unused or unsupported port");

  return false;
}

bool
CSPort::Send(uint8_t SendByte)
{
  // Dispatch
  switch(this->Type)
  {
    case PORT_BUFFERED: return this->BufferedSend(SendByte);
    case PORT_STATE:    return this->StateSend(SendByte);
    default: break;
  }

  // Seems unused or unsupported
  CLOG(LOG_WARNING, "Tried to send data to an unused or unsupported port");

  return false;
}

bool
CSPort::StateRecv(uint8_t *RetByte)
{
  // Just set the byte to the current state
  // It would be strange if RetByte is not provided, but I'll
  // allow such a situation
  if(RetByte)
    *RetByte = this->StateData;

  // Done
  return true;
}

bool
CSPort::StateSend(uint8_t SendByte)
{
  // Set the state
  this->StateData = SendByte;

  // Done
  return true;
}

bool
CSPort::BufferedRecv(uint8_t *RetByte)
{
  // Is anything in the buffer?
  if(this->BufferDataCount == 0)
    return false;

  // Pop one byte
  // If RetByte is unset, this is a simple
  // port drain that we do support
  if(RetByte)
    *RetByte = this->Buffer[0];

  this->BufferDataCount--;
  memmove(&this->Buffer[0], &this->Buffer[1], this->BufferDataCount);

  // Done
  return true;
}

bool
CSPort::BufferedSend(uint8_t SendByte)
{
  // Is the buffer large enough to containt another byte?
  if(this->BufferDataCount + 1 > (int)this->Buffer.size())
    return false;

  // Add the byte
  this->Buffer[this->BufferDataCount] = SendByte;
  this->BufferDataCount++;

  // Is this an interrupt-genereting port?
  if(this->IntNo != -1 && this->BufferDataCount >= this->IntDataCount)
  {
    // Issue an interrupt
    // This only adds the interrupt to the dispatcher, it 
    // does not freeze the system
    CS::IssueInterrupt(this->IntNo);
  }

  // Done
  return true;
}

bool
CSPort::SetInterrupt(int IntNo, int IntDataCount)
{
  // Interrupts are supported only on buffered ports
  // TODO: Think if we need interrupts for state ports
  if(this->Type != PORT_BUFFERED)
    return false;

  // Is this an unset?
  if(IntNo == -1)
    IntDataCount = -1;

  // Set
  this->IntNo        = IntNo;
  this->IntDataCount = IntDataCount;

  // Return
  return true;
}

int
CSPort::GetBufferSize() const
{
  // If it's not a buffered port, return -1
  if(this->Type != PORT_BUFFERED)
    return -1;

  // Just return the size
  return (int)this->Buffer.size();
}

bool
CSPort::SetBufferSize(int Size)
{
  // Only the buffered ports can reset the buffer size
  if(this->Type != PORT_BUFFERED)
    return false;

  // Please note that the data inside the buffer
  // will be truncated if the Size is smaller than
  // the current buffer - the most RECENT bytes
  // will be lost.

  // Resize, without error checks
  // In case of errors, the programmer will have to
  // figure something out
  this->Buffer.resize(Size);

  // Check if the data is to be truncated
  // (well it's acctually already truncated, but not
  // yet aware of it ;p)
  if(this->BufferDataCount > Size)
    this->BufferDataCount = Size;

  // Done
  return true;
}

bool 
CSPort::IsAssigned() const
{
  // Does this port have an owner?
  return (bool)(this->OwnerHw != NULL);
}

bool
CSPort::Assign(CHardware *OwnerHw)
{
  // Already assigned?
  if(this->IsAssigned())
    return false; // Sorry, already assigned

  // Assign
  this->OwnerHw = OwnerHw;

  // Done
  return true;  
}

bool
CSPort::SetType(enum PortType Type)
{
  // Just set the type
  // Why is it a switch, not a simple this->Type = Type?
  // Because:
  // 1. I want to make sure only supported types can be set
  // 2. Some types may require additional actions
  switch(Type)
  {
    // Disable port?
    case PORT_UNUSED:
      this->Type = Type;
      break;

    // Buffered!
    case PORT_BUFFERED:
      this->Type = Type;
      this->SetBufferSize(PORT_BUFFER_SIZE); // 1 byte is default
      this->BufferDataCount = 0; // No data in reset port!
      break;

    // State!
    case PORT_STATE:
      this->Type = Type;
      this->StateData = 0; // Default!
      break;

    // Unsupported?
    default:
      return false; // We don't change anything
  }

  // Done
  return true;
}

enum CSPort::PortType
CSPort::GetType() const
{
  // Just return the type
  return this->Type;
}


// ------------------------------------------------------------------
// CSInterrupt
// ------------------------------------------------------------------
CSInterrupt::CSInterrupt()
{
  this->IntNo      = -1;
  this->OwnerHw    = NULL;
}

CHardware*
CSInterrupt::GetOwner() const
{
  // Return owner
  return this->OwnerHw;
}

bool
CSInterrupt::SetIntNo(int IntNo)
{
  // Set interrupt number
  this->IntNo = IntNo;

  // Done
  return true;
}

bool
CSInterrupt::IsAssigned() const
{
  // Is it assigned ?
  return (bool)(this->OwnerHw != NULL);
}

bool
CSInterrupt::Assign(CHardware *OwnerHw)
{
  // Already assigned?
  if(this->IsAssigned())
    return false; // Sorry, already assigned

  // Assign
  this->OwnerHw = OwnerHw;

  // Done
  return true;
}

bool
CSInterrupt::Dispatch()
{
  // Any owner?
  if(!this->IsAssigned())
    return false; // Nope

  // Let the dispatcher handle it
  // Forward the return value
  return this->OwnerHw->InterruptDispatcherWrapper(this->IntNo, this->OwnerHw);
}

// ------------------------------------------------------------------
// CSystem
// ------------------------------------------------------------------
CSystem::CSystem()
{
  int i;

  // Setup interrupt numbers
  for(i = 0; i < INTERRUPT_COUNT; i++)
    this->Interrupts[i].SetIntNo(i);

  // Setup port numbers
  for(i = 0; i < PORT_COUNT; i++)
    this->Ports[i].SetPortNo(i);
}

// ------------------------------------------------------------------
// CS Namespace
// ------------------------------------------------------------------
namespace CS
{
  CSystem *System;
  unsigned long long CurrentCycle;
  LPVOID MainFiber;
  DWORD FPSTicks;
  unsigned long long FPSCycle;

  // Initialize computer system
  bool
  Init()
  {
    // Already initialized?
    if(System)
      return false;

    // Allocate
    System = new CSystem;

    // Convert self to fiber
    MainFiber = ConvertThreadToFiber(NULL);

    // Reset cycle counter
    CurrentCycle = 0;
    FPSCycle = 0;
    FPSTicks = GetTickCount();

    // Done
    return true;
  }

  bool
  IssueInterrupt(int IntNo)
  {
    // I really hope System was initialized ;)

    // Is IntNo sane?
    if(IntNo < 0 || IntNo >= INTERRUPT_COUNT)
      return false;

    // Is the interrupt assigned?
    // If so, dispatch it, and forward the result
    if(System->Interrupts[IntNo].IsAssigned())
      return System->Interrupts[IntNo].Dispatch();

    // Guess the interrupt wasn't assigned
    return false;
  }

  // Register stuff for hardware
  bool
  AcquirePort(int PortNo, CHardware *Hw)
  {
    // Is the port free?
    if(System->Ports[PortNo].IsAssigned())
    {
      // Already assigned, sorry
      return false;
    }

    // Assign
    return System->Ports[PortNo].Assign(Hw);
  }

  bool
  AcquireInterrupt(int IntNo, CHardware *Hw)
  {
    // Is this interrupt free?
    if(System->Interrupts[IntNo].IsAssigned())
    {
      // Already assigned, sorry
      return false;
    }

    // Assign
    return System->Interrupts[IntNo].Assign(Hw);
  }

  bool
  AddHardware(CHardware *Hw)
  {
    // Just add it to the list
    System->Hardware.push_back(Hw);

    // Done
    return true;
  }

  bool
  Step()
  {
    // Debug
    GLOG(LOG_SUPER_VERBOSE, "Cycle %.8x%.8x", 
        ((uint32_t*)&CurrentCycle)[1], 
        ((uint32_t*)&CurrentCycle)[0]);

    // Copy the hardware to a local vector and randomize it's content
    // This is done to simulate race conditions in bus access and other
    // stuff
    std::vector<CHardware *> LocalHw(System->Hardware);
    std::random_shuffle(LocalHw.begin(), LocalHw.end());

    // Default is OK
    bool RetValue = true;

    // Now, step the hardware
    int i, sz = (int)LocalHw.size();
    for(i = 0; i < sz; i++)
    {
      // Debug
      GLOG(LOG_SUPER_VERBOSE, "Steping \"%s\"", LocalHw[i]->GetName());

      // Step the hardware
      if(!LocalHw[i]->Step())
        RetValue = false;
    }

    // Add cycle
    CurrentCycle++;

    // Is it time to print the cycle counter ?
    if(GetTickCount() > FPSTicks + 1000)
    {
      // Calculate the MHZ
      unsigned long long CyclesPassed = CurrentCycle - FPSCycle;
      CyclesPassed /= 1000;

      float MHz = (float)CyclesPassed / 1000.0f;

      // Show it
      char MHzBuffer[256];
      sprintf(MHzBuffer, "Last cycle count per second: %f", MHz);

      SetConsoleTitle(MHzBuffer);
      
      // Store the current values
      FPSCycle = CurrentCycle;
      FPSTicks = GetTickCount();
    }

    // Now, the GUI stuff
    MSG Msg;
    while(PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE))
    {
      // Handle
      TranslateMessage(&Msg);
      DispatchMessage(&Msg);

      // Quit ?
      if(Msg.message == WM_QUIT)
        return false;
    }

    // Done
    return RetValue;
  }

};

