#pragma once
#if _WIN32_WINNT < 0x0400
#  undef _WIN32_WINNT
#  define _WIN32_WINNT 0x0400
#endif
#include <vector>
#include <stdint.h>
#include <windows.h>
#include "CHardware.h"
#include "CLog.h"

const int PORT_COUNT       = 0x1000;
const int INTERRUPT_COUNT  = 0x100;
const int PORT_BUFFER_SIZE = 1;

// Architectural notice:
// This architecture assumes that the port bus is common to
// every hardware piece on the board. Additionally, the port
// numbers are common on every hardware, and do not need to
// be asigned individually on both ends.
// Also, the interrupt bus is common for every hardware,
// including the interrupt numbers.
// Please note, that this is a simplification, because on
// a normal system each CPU/uC has it's own interrups scheme
// and ports, and ports need to be connected between each 
// piece before they can be used for communication.

// Declare hardware class
class CHardware;

// CSPort
// Port has a PORT_BUFFER_SIZE buffer at start, PORT_BUFFER_SIZE is
// 1 by default.
// The buffer size can be changed to any size greater than 0.
// In case the port has an interrupt configured, then 
// an interrupt is signaled when the buffer has at least
// IntDataCount bytes.
// If the buffer size is exceeded, new bytes are not
// added to it.
class CSPort : public CLog
{
  public:
    enum PortType
    {
      PORT_UNUSED,   // Unused port - turned off, does not receive data,
                     // does not give data, does not generate interrupts
                     // basically, it's off

      PORT_BUFFERED, // Buffered port - it has a finite buffer, that can be
                     // written to using Send() and read from using Recv()
                     // If the buffer is drained by Recv(), nothing is returned
                     // until the buffer gets (at least partly, as in 1 byte)
                     // populated by the Send() command.
                     // Default buffer size is 1 byte, and the buffer is empty

      PORT_STATE     // State port - the port returns always one byte of data
                     // This byte can be set using Send() routine (the last
                     // byte sent is considered the current state of the port),
                     // and it is always returned on Recv() command (i.e. it's
                     // not drained, since it's not a buffer)
                     // The state is, by default, set to zero
    };

  protected:
    // Port type
    enum PortType Type;

    // Port number (for debug/warning messages)
    int PortNo;

    // Buffered port data
    std::vector<uint8_t> Buffer;
    int BufferDataCount;
    bool BufferedSend(uint8_t SendByte);
    bool BufferedRecv(uint8_t *RetByte);
    
    // State port data
    uint8_t StateData;
    bool StateSend(uint8_t SendByte);
    bool StateRecv(uint8_t *RetByte);

    // Owner and signal interrupts
    CHardware *OwnerHw;
    int     IntNo;
    int     IntDataCount;

    // Get name
    char PortName[20]; // PortBus($1234)
    const char* GetName() const;    

  public:
    CSPort();
    bool IsDataPresent() const;
    bool Recv(uint8_t *RetByte);
    bool Send(uint8_t SendByte);
    bool SetInterrupt(int IntNo, int IntDataCount);
    int GetBufferSize() const;
    bool SetBufferSize(int Size);
    bool IsAssigned() const;
    bool Assign(CHardware *OwnerHw);
    CHardware* GetOwner() const;
    bool SetType(enum PortType Type);
    enum PortType GetType() const;

    // Internal use
    void SetPortNo(int PortNo);
};

// Interrupts
// An interrupts dispatcher will call the interrupt handler
// function in the given piece of hardware
// The hardware itself is responsible for queueing the
// interrupts, selecting priorities, etc
// (the CSHardware class does it by default)

class CSInterrupt
{
  protected:
    // Interrupt dispatcher type
    int IntNo;
    CHardware *OwnerHw;

  public:
    CSInterrupt();
    bool SetIntNo(int IntNo); // Private
    bool IsAssigned() const;
    bool Assign(CHardware *OwnerHw);
    bool Dispatch();
    CHardware* GetOwner() const;
};


// CSystem
// A whole system class - it has a list of hardware, a list of
// CPUs with frequences, etc; it also maintaines the ports and 
// interrupts
class CSystem
{
  public:
    // Port and Interrupt Bus
    CSPort      Ports[PORT_COUNT];
    CSInterrupt Interrupts[INTERRUPT_COUNT];

    // Hardware elements attached to the system
    std::vector<CHardware *> Hardware;

    CSystem();
};

namespace CS
{
  extern CSystem *System;
  extern unsigned long long CurrentCycle;
  extern LPVOID MainFiber;

  bool Init();
  bool IssueInterrupt(int IntNo);

  // Accquire (reserve) Interrupt and Port by a 
  // specific part of hardware
  bool AcquirePort(int PortNo, CHardware *Hw);
  bool AcquireInterrupt(int IntNo, CHardware *Hw);

  // Add hardware element
  bool AddHardware(CHardware *Hw);

  // Step system
  bool Step();

};

// Usable macros
#define UNUSED (void)


