#include "CHWCRT.h"

// ------------------------------------------------------------------
// CRT public methods
// ------------------------------------------------------------------
CHWCRT::CHWCRT()
{
  // Load the TGA image
  ImageTGA32 tga;
  if(!tga.Load("gfx/szyba.tga"))
    return;

  // Setup images
  // Hardcode them for now
  this->ImgGlass.Setup(512, 512); // TODO Constantize this value
  this->ImgBuffer.Setup(512, 512);
  this->ImgOutput.Setup(512, 512);

  // Convert the tga to the float image
  // TODO Place this in a function
  int i;
  RGBAf *pOut = &this->ImgGlass.data[0];
  uint8_t *pIn = &tga.Data[0];
  for(i = 0; i < 512 * 512 * 4; i += 4, pOut++, pIn += 4)
  {
    pOut->b = (float)pIn[0] / 255.0f;
    pOut->g = (float)pIn[1] / 255.0f;
    pOut->r = (float)pIn[2] / 255.0f;
    pOut->a = (float)pIn[3] / 255.0f;
  }

  // Create the ports
  int BufferedPortsNo[] = 
  {
    PORT_IN_DATA_R, 
    PORT_IN_DATA_G, 
    PORT_IN_DATA_B, 
    PORT_IN_DATA_MONO
  };

  int StatusPortsNo[] =
  {
    PORT_STATE_X,
    PORT_STATE_Y_LO,
    PORT_STATE_Y_HI,
    PORT_IN_DEFAULT_R,
    PORT_IN_DEFAULT_G,
    PORT_IN_DEFAULT_B,
    PORT_IN_DEFAULT_MONO,
    PORT_HALT,
    PORT_MODE
  };

  // Setup buffered ports
  for(i = 0; i < (int)(sizeof(BufferedPortsNo) / sizeof(*BufferedPortsNo)); i++)
  {
    // Acquire
    int PortNo = BufferedPortsNo[i];
    CS::AcquirePort(PortNo, this);

    // Set as buffered port
    CS::System->Ports[PortNo].SetType(CSPort::PORT_BUFFERED);
    CS::System->Ports[PortNo].SetBufferSize(512); // 512 bytes
  }

  for(i = 0; i < (int)(sizeof(StatusPortsNo) / sizeof(*StatusPortsNo)); i++)
  {
    // Acquire
    int PortNo = StatusPortsNo[i];
    CS::AcquirePort(PortNo, this);

    // Set as state port
    CS::System->Ports[PortNo].SetType(CSPort::PORT_STATE);
  }

  // Setup initial settings
  this->CurrentCycle = this->LastPacketCycle = 0;
  this->CurrentInputFormat = FMT_MONO;
  this->ChangeResolution(RES_256_256);
  this->ResetFrame();

  // Set the mode port
  CS::System->Ports[PORT_MODE].Send(
      (uint8_t)this->CurrentInputFormat |
      (uint8_t)this->CurrentResolution
      );

  // Done

  // Register the window class
  if(!this->WndRegisterClass())
  {
    // XXX Destroy something?
    return;
  }

  // Show the window
  if(!this->WndCreate())
  {
    // XXX Destroy something?
    return;
  }
}

CHWCRT::~CHWCRT()
{
  // Do nothing yet
}

const char *
CHWCRT::GetName() const
{
  // Return the name
  return "HW_CRT";
}

bool
CHWCRT::InterruptDispatcher(int IntNo)
{
  // Basically, do nothing
  UNUSED IntNo;

  // Well, the CRT does not have any interrupts
  return true;
}

bool
CHWCRT::Step()
{
  int i;
  RGBi Colors[32];

  // Are we in halt mode ?
  // (Make this check only on EOLN
  if(this->CurrentStateX == 0)
  {
    // Get the state
    uint8_t Halt = 0;
    CS::System->Ports[PORT_HALT].Recv(&Halt);

    // Halt at the end of frame ?
    // Halt at the end of line ?
    if((Halt == 0x01 && this->CurrentStateY == 0) ||
       (Halt == 0xff))
    {
      // Yep, halt execution
      // The cycles are NOT incremented in halt mode, so there
      // still is the CyclesPerPacket lag after restarting the 
      // rendering
      return true;  
    }
  }



  // Not in halt state (any more?)
  // Should we draw a pixel ?
  if(this->CurrentCycle >= this->LastPacketCycle + this->CyclesPerPacket)
  {
    // Get the settings
    uint8_t Mode = 0;
    CS::System->Ports[PORT_MODE].Recv(&Mode);

    // Setting the resolution won't acctually change it now. It
    // will be changed at the end of the line.
    this->CurrentResolution  = (enum ModeResolution)(Mode & 0x0f);
    this->CurrentInputFormat = (enum ModeInputFormat)(Mode & 0xf0);

    // Get the data
    this->GetColors(Colors);
    for(i = 0; i < 32; i++)
    {
      // Draw pixel
      this->DrawPixel(Colors[i].r, Colors[i].g, Colors[i].b);

      // Advance
      this->CurrentBufferX += this->CurrentPixelW;
      this->CurrentStateX++;

      // Don't need to check if we go over the X, since
      // it's all calculated, and packets are aligned :)
    }

    // Advance
    this->AdvancePacket();

    // Mark time
    this->LastPacketCycle = this->CurrentCycle;
  }

  // Advance cycle
  this->CurrentCycle++;

  // Done
  return true;
}

// ------------------------------------------------------------------
// CRT GUI stuff
// ------------------------------------------------------------------
bool 
CHWCRT::WndRegisterClass()
{
  // Variables
  WNDCLASSEX wc;

  // Setup the class
  memset(&wc, 0, sizeof(wc));
  wc.cbSize        = sizeof(WNDCLASSEX);
  wc.style         = 0;
  wc.lpfnWndProc   = this->WndProc;
  wc.hInstance     = GetModuleHandle(0);
  wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION); // TODO an icon
  wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION); // TODO an icon
  wc.hCursor       = LoadCursor(NULL, IDC_ARROW); 
  wc.hbrBackground = (HBRUSH)(COLOR_WINDOW);
  wc.lpszMenuName  = NULL; // No menu
  wc.lpszClassName = "HWCRT_Display"; // TODO constantize this

  // Try to register the class
  if(!RegisterClassEx(&wc))
  {
    // Log
    CLOG(LOG_CRITICAL, "Could not register window class (LastError=%i)", GetLastError());

    // Huh
    return false;
  }
  
  // Done
  return true;
}

bool
CHWCRT::WndCreate()
{
  // Style
  DWORD Style = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;

  // Calc window size
  RECT rc = { 0, 0, 512, 512 }; // TODO Constantize this
  AdjustWindowRect(&rc, Style, FALSE);
  
  // Create the window
  HWND hwnd;
  hwnd = CreateWindowEx(
      0,
      "HWCRT_Display",
      "HWCRT_Display",
      Style | WS_VISIBLE,
      CW_USEDEFAULT, CW_USEDEFAULT,
      rc.right - rc.left,
      rc.bottom - rc.top,
      (HWND)NULL,
      (HMENU)NULL,
      GetModuleHandle(0),
      (LPVOID)this
    );

  // All OK?
  if(!hwnd)
  {
    // Log
    CLOG(LOG_CRITICAL, "Could not create window (LastError=%i)", GetLastError());

    // Huh
    return false;
  }

  // Set
  this->Wnd = hwnd;

  // Done
  return true;
}

LRESULT CALLBACK 
CHWCRT::WndProc(HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
  // Get the class instance
  CHWCRT *Self = NULL;

  // WM_CREATE is a special case here
  if(Msg == WM_CREATE)
  {
    // Get the init structure
    CREATESTRUCT *data = (CREATESTRUCT*)lParam;

    // Set Self
    Self = (CHWCRT*)data->lpCreateParams;
    
    // Store it
    SetWindowLongPtr(Wnd, GWLP_USERDATA, (LONG_PTR)Self);
  }
  else
  {
    // Retrieve the pointer
    Self = (CHWCRT*)GetWindowLongPtr(Wnd, GWLP_USERDATA);
  }

  // Message? What message?
  switch(Msg)
  {
    // Generally, we need only to cover a few messages here
    
    case WM_PAINT: return Self->WndMsg_Paint(Wnd);

    case WM_CLOSE: 
      PostQuitMessage(0);
      // Pass

    // The rest of messages are handled by DefWindowProc
    default:
      return DefWindowProc(Wnd, Msg, wParam, lParam);
  }

  // Done
  // Unreachable?
  // return 0;
}

int
CHWCRT::WndMsg_Paint(HWND Wnd)
{
  PAINTSTRUCT ps;
  
  // Get the context / begin paint
  HDC hdc = BeginPaint(Wnd, &ps);

  // Copy the part of the buffer required
  static const BITMAPINFOHEADER bmhdr =
  {
    sizeof(BITMAPINFOHEADER),
    512, // TODO Constantize
    -512,
    1,
    32,
    BI_RGB,
    0,0,0,0,0
  };

  // Cache values
  int x = ps.rcPaint.left;
  int y = ps.rcPaint.top;
  int w = ps.rcPaint.right - ps.rcPaint.left;
  int h = ps.rcPaint.bottom - ps.rcPaint.top;

  // Draw
  StretchDIBits(hdc, x, y, w, h, x, y, w, h, 
      (const VOID*)&this->ImgOutput.data[0],
      (BITMAPINFO*)&bmhdr,
      DIB_RGB_COLORS,
      SRCCOPY
  );

  // Done
  EndPaint(Wnd, &ps);

  // Done
  return 0;
}

// ------------------------------------------------------------------
// CRT private image handling methods and classes
// ------------------------------------------------------------------

bool
CHWCRT::ImgMergeAndConvert(Image24i *Out, Image32f *In1, Image32f *In2)
{
  // Check if images are same sizes
  if(In1->w != In2->w ||
     In1->h != In2->h ||
     Out->w != In2->w ||
     Out->h != In2->h)
    return false; // Not

  // Let us get a pointer!
  RGBAf *pIn1 = &In1->data[0];
  RGBAf *pIn2 = &In2->data[0];
  RGBi  *pOut = &Out->data[0];
  int i, sz = In1->w * In1->h;

  for(i = 0; i < sz; i++, pIn1++, pIn2++, pOut++)
  {
    float Alpha         = pIn1->a;
    float OneMinusAlpha = 1.0f - pIn1->a;
    pOut->r = (uint8_t)((pIn1->r * Alpha + pIn2->r * OneMinusAlpha) * 255.0);
    pOut->g = (uint8_t)((pIn1->g * Alpha + pIn2->g * OneMinusAlpha) * 255.0);
    pOut->b = (uint8_t)((pIn1->b * Alpha + pIn2->b * OneMinusAlpha) * 255.0);
  }

  // Done
  return true;
}

bool
CHWCRT::Image24i::Setup(int w, int h)
{
  // Just set the variables
  this->w = w;
  this->h = h;
  this->data.resize(w * h); // TODO how to check resize?

  // Done I guess
  return true;
}

bool
CHWCRT::Image32f::Setup(int w, int h)
{
  // Just set the variables
  this->w = w;
  this->h = h;
  this->data.resize(w * h);

  // Done I guess
  return true;
}

// ------------------------------------------------------------------
// CRT inner workings
// ------------------------------------------------------------------
void
CHWCRT::ChangeResolution(enum ModeResolution NewResolution)
{
  // Let's set up the new resolution
  switch(NewResolution)
  {
    case RES_512_512: 
      this->CurrentResolution = NewResolution;
      this->CurrentPixelW = this->CurrentPixelH = 1;
      this->CyclesPerPacket = 8;
      break;

    case RES_256_256: 
      this->CurrentResolution = NewResolution;
      this->CurrentPixelW = this->CurrentPixelH = 2;
      this->CyclesPerPacket = 32;      
      break;

    case RES_128_128: 
      this->CurrentResolution = NewResolution;
      this->CurrentPixelW = this->CurrentPixelH = 4;
      this->CyclesPerPacket = 128;      
      break;

    case RES_64_64: 
      this->CurrentResolution = NewResolution;
      this->CurrentPixelW = this->CurrentPixelH = 8;
      this->CyclesPerPacket = 512;      
      break;

    default:
      CLOG(LOG_WARNING, "Unknown resolution selected (%.2x)", 
          (uint8_t)this->CurrentResolution);
      break;
  }
}

void
CHWCRT::ResetFrame()
{
  // Guess we zero evetything
  this->CurrentBufferX = this->CurrentBufferY = 0;
  this->CurrentStateX  = this->CurrentStateY  = 0;

  // And recalculate everything
  this->ImgMergeAndConvert(
      &this->ImgOutput,
      &this->ImgGlass,
      &this->ImgBuffer
  );

  // Now we flush the window
  InvalidateRect(this->Wnd, NULL, false);
  UpdateWindow(this->Wnd);  
}

void
CHWCRT::AdvancePacket()
{
  // Is this a new line?
  if(this->CurrentBufferX >= 512)
    this->AdvanceLine();

  // Done
}

void
CHWCRT::AdvanceLine()
{
  // Reset X
  this->CurrentBufferX = 0;
  this->CurrentStateX  = 0;

  // Reset Y
  this->CurrentStateY++;
  this->CurrentBufferY += this->CurrentPixelH;

  // Let's change the pixel setting
  this->ChangeResolution(this->CurrentResolution);

  // Should we reset frame ?
  if(this->CurrentBufferY >= 512)
    this->ResetFrame();

  // Done
}

void
CHWCRT::DrawPixel(uint8_t iR, uint8_t iG, uint8_t iB)
{
  // Convert colors
  float r = (float)iR / 255.0f;
  float g = (float)iG / 255.0f;
  float b = (float)iB / 255.0f;

  // Calc the boundery
  int bX = this->CurrentBufferX + this->CurrentPixelW;
  int bY = this->CurrentBufferY + this->CurrentPixelH;
  //bX = bX > 512 ? 512 : bX; // this is not needed due to packet aligment
  bY = bY > 512 ? 512 : bY;

  // Draw
  int i, j;
  RGBAf *pOutRow = &ImgBuffer.data[this->CurrentBufferX + this->CurrentBufferY * 512];
  for(j = this->CurrentBufferY; j < bY; j++, pOutRow += 512)
  {
    // Pixels
    RGBAf *pOutPixel = pOutRow;

    for(i = this->CurrentBufferX; i < bX; i++, pOutPixel++)
    {
      // Draw
      pOutPixel->r = r;
      pOutPixel->g = g;
      pOutPixel->b = b;
    }
  }

  // Done
}

void
CHWCRT::GetColors(RGBi Colors[32])
{
  // What kinda input?
  switch(this->CurrentInputFormat)
  {
    case FMT_MONO: GetColorsMono(Colors); break;
    case FMT_COLOR_1_BPP: GetColors1bpp(Colors); break;
    case FMT_COLOR_2_BPP: GetColors2bpp(Colors); break;
    case FMT_COLOR_4_BPP: GetColors4bpp(Colors); break;
    case FMT_COLOR_8_BPP: GetColors8bpp(Colors); break;
    default: break; // WTF ?
  }
}

uint8_t
CHWCRT::GetByteWithDefault(int Port, int DefaultPort)
{
  uint8_t Ret = 0;

  // Try to get the data from the port
  if(!CS::System->Ports[Port].Recv(&Ret))
  {
    // Fall back to default state port
    CS::System->Ports[DefaultPort].Recv(&Ret);
  }

  // Return
  return Ret;
}

void
CHWCRT::GetColors8bpp(RGBi Colors[32])
{
  // For each pixel...
  int i;
  for(i = 0; i < 32; i++)
  {
    // Get the color values for R G an B
    Colors[i].r = GetByteWithDefault(PORT_IN_DATA_R, PORT_IN_DEFAULT_R);
    Colors[i].g = GetByteWithDefault(PORT_IN_DATA_G, PORT_IN_DEFAULT_G);
    Colors[i].b = GetByteWithDefault(PORT_IN_DATA_B, PORT_IN_DEFAULT_B);
  }
}

void
CHWCRT::GetColors2bpp(RGBi Colors[32])
{
  // Values
  uint8_t PalValues[4] = { 0, 84, 168, 255 };

  // For each pixel...
  int i;
  uint8_t ByteR = 0, ByteG = 0, ByteB = 0;
  int BitsLeft = 0;
  for(i = 0; i < 32; i++, BitsLeft-=2,
      ByteR = (ByteR >> 2),
      ByteG = (ByteG >> 2),
      ByteB = (ByteB >> 2))
  {
    // Get byte ?
    if(BitsLeft == 0)
    {
      // Get it!
      ByteR = GetByteWithDefault(PORT_IN_DATA_R, PORT_IN_DEFAULT_R);
      ByteG = GetByteWithDefault(PORT_IN_DATA_G, PORT_IN_DEFAULT_G);
      ByteB = GetByteWithDefault(PORT_IN_DATA_B, PORT_IN_DEFAULT_B);
      BitsLeft = 8;
    }

    // Get the color values for R G an B
    Colors[i].r = PalValues[ByteR & 3];
    Colors[i].g = PalValues[ByteG & 3];
    Colors[i].b = PalValues[ByteB & 3];
  }
}

void
CHWCRT::GetColors4bpp(RGBi Colors[32])
{
  // Values
  uint8_t PalValues[16] = {
    0, 17, 34, 51,
    68, 85, 102, 119,
    136, 153, 170, 187,
    204, 221, 238, 255
  };

  // For each pixel...
  int i;
  uint8_t ByteR = 0, ByteG = 0, ByteB = 0;
  int BitsLeft = 0;
  for(i = 0; i < 32; i++, BitsLeft-=4,
      ByteR = (ByteR >> 4),
      ByteG = (ByteG >> 4),
      ByteB = (ByteB >> 4))
  {
    // Get byte ?
    if(BitsLeft == 0)
    {
      // Get it!
      ByteR = GetByteWithDefault(PORT_IN_DATA_R, PORT_IN_DEFAULT_R);
      ByteG = GetByteWithDefault(PORT_IN_DATA_G, PORT_IN_DEFAULT_G);
      ByteB = GetByteWithDefault(PORT_IN_DATA_B, PORT_IN_DEFAULT_B);
      BitsLeft = 8;
    }

    // Get the color values for R G an B
    Colors[i].r = PalValues[ByteR & 0xf];
    Colors[i].g = PalValues[ByteG & 0xf];
    Colors[i].b = PalValues[ByteB & 0xf];
  }
}

void
CHWCRT::GetColors1bpp(RGBi Colors[32])
{
  // For each pixel...
  int i;
  uint8_t ByteR = 0, ByteG = 0, ByteB = 0;
  int BitsLeft = 0;
  for(i = 0; i < 32; i++, BitsLeft--,
      ByteR = (ByteR >> 1),
      ByteG = (ByteG >> 1),
      ByteB = (ByteB >> 1))
  {
    // Get byte ?
    if(BitsLeft == 0)
    {
      // Get it!
      ByteR = GetByteWithDefault(PORT_IN_DATA_R, PORT_IN_DEFAULT_R);
      ByteG = GetByteWithDefault(PORT_IN_DATA_G, PORT_IN_DEFAULT_G);
      ByteB = GetByteWithDefault(PORT_IN_DATA_B, PORT_IN_DEFAULT_B);
      BitsLeft = 8;
    }

    // Get the color values for R G an B
    Colors[i].r = (ByteR & 1) * 0xff;
    Colors[i].g = (ByteG & 1) * 0xff;
    Colors[i].b = (ByteB & 1) * 0xff;
  }
}

void
CHWCRT::GetColorsMono(RGBi Colors[32])
{
  // Colors
  RGBi PxColors[2] = {
    { 0, 0, 0, 0 },      // true black
    { 0xb0,0xff,0x95,0}  // old monitor green
  };

  // For each pixel...
  int i;
  uint8_t Byte = 0;
  int BitsLeft = 0;
  for(i = 0; i < 32; i++, BitsLeft--, Byte = (Byte >> 1))
  {
    // Get byte ?
    if(BitsLeft == 0)
    {
      // Get it!
      Byte = GetByteWithDefault(PORT_IN_DATA_MONO, PORT_IN_DEFAULT_MONO);
      BitsLeft = 8;
    }

    // Get the color values for R G an B
    Colors[i] = PxColors[Byte & 1];
  }
}


