#pragma once
#include <windows.h>
#include <vector>
#include "CHardware.h"
#include "TGALoader.h"

class CHWCRT : public CHardware
{
  protected:
    HWND Wnd; // Window
    static LRESULT CALLBACK WndProc(HWND Wnd, UINT Msg, WPARAM wParam, LPARAM lParam);
    bool WndRegisterClass();
    bool WndCreate();

    int  WndMsg_Paint(HWND Wnd);

    // RGBA float image
    struct RGBAf
    {
      float r,g,b,a;
    };

    struct Image32f
    {
      std::vector<RGBAf> data;
      int w, h;

      bool Setup(int w, int h);      
    };

    // RGB uint8 image
#pragma pack(push, 1)
    struct RGBi
    {
      uint8_t b,g,r,unused;
    };
#pragma pack(pop)

    struct Image24i
    {
      std::vector<RGBi> data;
      int w, h;

      bool Setup(int w, int h);
    };

    Image32f ImgGlass;
    Image32f ImgBuffer;
    Image24i ImgOutput;

    // Alpha of Image 1 is used only
    // Res = In1.Alpha * In1.Colors + (1.0 - In1.Alpha) * In2.Colors
    bool ImgMergeAndConvert(Image24i *Out, Image32f *In1, Image32f *In2);


    // CRT stuff

    // The X and Y of the upper left corner of the next pixel to draw
    int CurrentBufferX, CurrentBufferY;

    // The Width and Height of the pixel in the currently selected mode
    int CurrentPixelW, CurrentPixelH;

    // The state X and Y for the gfx card to get
    int CurrentStateX, CurrentStateY;

    // Cycles
    uint64_t CurrentCycle, LastPacketCycle;
    uint64_t CyclesPerPacket;

    // Modes
    enum ModeResolution
    {
      RES_512_512 = 1,
      RES_256_256 = 2,
      RES_128_128 = 4,
      RES_64_64   = 8
    };

    enum ModeInputFormat
    {
      FMT_MONO = 0x0,
      FMT_COLOR_1_BPP = 0x10,
      FMT_COLOR_2_BPP = 0x20,
      FMT_COLOR_4_BPP = 0x40,
      FMT_COLOR_8_BPP = 0x80
    };

    enum ModeResolution  CurrentResolution;
    enum ModeInputFormat CurrentInputFormat;

    // Setup everything to draw the next frame
    void ResetFrame();

    // Change resolution
    void ChangeResolution(enum ModeResolution NewResolution);

    // Advance
    void AdvancePacket(); // Loop calls this. This might call
    void AdvanceLine();   // this. And this might call ResetFrame
    void DrawPixel(uint8_t iR, uint8_t iG, uint8_t iB);
    void GetColors8bpp(RGBi Colors[32]);
    void GetColors4bpp(RGBi Colors[32]);
    void GetColors2bpp(RGBi Colors[32]);
    void GetColors1bpp(RGBi Colors[32]);
    void GetColorsMono(RGBi Colors[32]);
    void GetColors(RGBi Colors[32]);
    uint8_t GetByteWithDefault(int Port, int DefaultPort);
     
  public:
    // --------------------------------------------------------------
    // CRT Ports
    // --------------------------------------------------------------
    // The pixels are emited to the screen in packets of 32 pixels. Each
    // emit takes 8 cycles. So, the X status ports change every 32 pixels
    // (so it's either 16, 8, 4 or 2 times per line, depending on the
    // resolution selected).
    // The width and height of the pixels depend on the current resolution.
    //
    // When the X reaches end of the line, there is no delay before
    // the X is reset and Y is increased.
    //
    // At the end of the frame, there is an 8 cycle delay, before both
    // X and Y are reset.
    //
    // The drawing can be stoped at the end of the line or end of the frame.
    // To do thism issue a command to the HALT port (see below).
    //
    // The total frame cycle count is constant, and does not depend on the
    // resolution. It's 65544 cycles per frame.
    //
    // The initial mode is 512 x 512 x MONO 1 bpp
    //
    // The following resolutions are available:
    // Mode    X     Y   pixel size    cycles per packet     packets  note
    //    1  512 x 512        1 x 1                    8        8192  hi-res, good quality, high bandwidth required
    //    2  256 x 256        2 x 2                   32        2048
    //    4  128 x 128        4 x 4                  128         512 
    //    8   64 x  64        8 x 8                  512         128 lo-res, low bandwidth required
    //
    // As far as input data format modes are concerned:
    // Mode      Input Format             Max colors   Description
    // 0x00      Octets (8 px), 1 bpp              2   Monochromatic, 8 px at a time, _MONO ports
    // 0x10      Octets (8 px), 3*1 bpp            8   Color, 8 px at a time, _R/G/B ports
    // 0x20      Octets (4 px), 3*2 bpp           64   Color, 4 px at a time, _R/G/B ports
    // 0x40      Octets (2 px), 3*4 bpp         4096   Color, 2 px at a time, _R/G/B ports
    // 0x80      Octets (1 px), 3*8 bpp     16777216   Color, 1 px at a time, _R/G/B ports
    //
    // To set the mode combine (OR) the resolution mode with the input format mode, and set
    // the value on the below port. The resolution is changed from the next line. The input
    // The input format is changed from the next pixel packet (32 pixels).
    static const int PORT_MODE        = 0x020E;

    // The current X and Y of the thrown electron package
    // If you draw a line in mode 0x?1, and then switch to mode 0x?8 for the next line
    // And after it switch back to 0x?1, the Y counter will still be "2" at this time,
    // since the line height is not taken into account while drawing the lines.
    // X is in packets
    static const int PORT_STATE_X    = 0x0200;
    static const int PORT_STATE_Y_LO = 0x0201;
    static const int PORT_STATE_Y_HI = 0x0202;

    // The CRT uses the color of the R G and B values
    // for the current pixel from the below ports
    // If the DATA_R/G/B port contains any data, then the
    // data from it is used, otherwise the values from the
    // status ports DEFAULT_R/G/B are used.
    // The data buffers are 512 BYTES long.
    // The 0x00 mode (monochrome) uses the MONO port, that is
    // also buffered with 512 bytes long buffer.
    static const int PORT_IN_DEFAULT_R = 0x0210;
    static const int PORT_IN_DEFAULT_G = 0x0211;
    static const int PORT_IN_DEFAULT_B = 0x0212;
    static const int PORT_IN_DEFAULT_MONO = 0x0213;

    static const int PORT_IN_DATA_R = 0x0214;
    static const int PORT_IN_DATA_G = 0x0215;
    static const int PORT_IN_DATA_B = 0x0216;
    static const int PORT_IN_DATA_MONO = 0x0217;

    // The HALT port accepts 3 states / commands:
    // 00 - no halt
    // 01 - halt at end of frame
    // FF - halt at end of line
    static const int PORT_HALT = 0x020F;

    CHWCRT();
    ~CHWCRT();
    const char *GetName() const;
    bool InterruptDispatcher(int IntNo);
    bool Step();
};


