#include "CHWMemory.h"

CHWMemory::CHWMemory()
{
  // Log
  CLOG(LOG_VERBOSE, "Acquiering port %.4x", PORT_REQUESTS);

  // Reserve ports and interrupts
  CS::AcquirePort(PORT_REQUESTS, this);

  // Set buffer size for the port
  // The buffer is set to containt at max 16 commands, all the rest
  // is forgoten. One command per cycle.
  CS::System->Ports[PORT_REQUESTS].SetType(CSPort::PORT_BUFFERED);
  CS::System->Ports[PORT_REQUESTS].SetBufferSize(MEMORY_COMMAND_SIZE * 0x10);
}

const char *
CHWMemory::GetName() const
{
  return "HW_Memory";
}

bool
CHWMemory::InterruptDispatcher(int IntNo)
{
  // Whaat ? I has no interrupt
  (void) IntNo; // Unused
  return false;
}

bool
CHWMemory::Step()
{
  // Try to fill the command
  while(this->CurrentCommandBytesReceived != MEMORY_COMMAND_SIZE)
  {
    // Read the byte
    uint8_t Byte;
    if(!CS::System->Ports[PORT_REQUESTS].Recv(&Byte))
      break;

    // Write the byte to the buffer
    this->CurrentCommand.Raw[this->CurrentCommandBytesReceived++] = Byte;    
  }

  // A new command ?
  if(this->CurrentCommandBytesReceived != MEMORY_COMMAND_SIZE)
  {
    // Not yet. Either the data has not yet fully arrived
    // or we have nothing to do.
    // A wasted cycle.
    return true;
  }

  // Reset the CurrentCommandBytesReceived counter - we 
  // already know the data is waiting
  this->CurrentCommandBytesReceived = 0;

  // Execute command
  switch(this->CurrentCommand.Decoded.Opcode)
  {
    case OPCODE_READ:  return this->HandleRead();
    case OPCODE_WRITE: return this->HandleWrite();
    default:           break; // What'ed just happend?
  }
  
  // Done!
  return true;
}

bool
CHWMemory::HandleRead()
{
  // Let's read (if the dst port buffer is not enough, we don't care!)
  uint32_t Addr = this->CurrentCommand.Decoded.Address;
  uint32_t Size = this->CurrentCommand.Decoded.DataSize;
  uint16_t Port = this->CurrentCommand.Decoded.DstPort;

  // Debug
  CLOG(LOG_VERBOSE, "Read command: Addr=%.4x Size=%.4x Port=%.4x, PortOwner=\"%s\"", 
      Addr, Size, Port, CS::System->Ports[Port].IsAssigned() ? CS::System->Ports[Port].GetOwner()->GetName() : "port not assigned"
      );

  uint32_t i;
  for(i = Addr; i < Addr + Size; i++)
  {
    // Get byte
    uint8_t Byte = i < (uint32_t)MEMORY_TOTAL ? this->Data[i] : 0x00;
    
    // Just write data to buffer
    CS::System->Ports[Port].Send(Byte);
  }

  // Done!
  return true;
}

bool
CHWMemory::HandleWrite()
{
  uint32_t Addr = this->CurrentCommand.Decoded.Address;
  uint32_t Size = this->CurrentCommand.Decoded.DataSize;
  uint16_t Port = this->CurrentCommand.Decoded.DstPort; // This is src port
  uint32_t i;
  for(i = Addr; i < Addr + Size; i++)
  {
    // Get byte from port
    uint8_t Byte;
    if(!CS::System->Ports[Port].Recv(&Byte))
      Byte = 0x00;
    
    // Write data
    if(i < (uint32_t)MEMORY_TOTAL)
      this->Data[i] = Byte;    
  }

  // Done!
  return true;
}

