#include "CHardware.h"

const int MEMORY_TOTAL = 0x10000;
const int MEMORY_COMMAND_SIZE = 6;

// CHWMemory Ports & Protocol:
// Port: PORT_REQUESTS == 0x0010
//
// Request protocol format:
// All 2-byte values are little endian
//
// READ:
// [1 byte] OPCODE_READ == 0x00
// [2 byte] Address of data
// [1 byte] Data size (must be <= 16 bytes)
// [2 byte] Port to write data to (buffer must be at least 16 bytes)
//
// WRITE:
// [1 byte] OPCODE_WRITE == 0x01
// [2 byte] Address of data
// [1 byte] Data size (must be <= 16 bytes)
// [2 byte] Port to read data from
//
// E.g.:
// 00 00 01 10 02 02 - Read 16 bytes from 0100h to port 0202h
// 01 34 12 01 01 02 - Write 1 byte to 1234h from port 0201h
//
// Notes:
// If read address of data + data size exceed memory limit, 00 is 
// returned as the left over bytes.
// If write source port does not have enough data, 00 value is
// used.
// If the address of write data + data size exceed memory limit,
// it is ignored, but data is drained from the port and it apears
// later as dust on the main chip.
// The memory controler executes one command per cycle, so no more
// than 16 bytes written or read can be done in one cycle.
//

class CHWMemory : public CHardware
{
  protected:
    uint8_t Data[MEMORY_TOTAL];
    
    // Command temporary buffer
    #pragma pack(push,1)
    union
    {
      uint8_t Raw[MEMORY_COMMAND_SIZE];
      struct
      {
        uint8_t  Opcode;
        uint16_t Address; // Address!!!
        uint8_t  DataSize;
        uint16_t DstPort; // this is source in Write
      } Decoded;
    } CurrentCommand;
    #pragma pack(pop)

    // Byte counter for the command
    int CurrentCommandBytesReceived;

    // Both these use the above structure and buffer
    bool HandleRead();
    bool HandleWrite();

  public:
    static const int PORT_REQUESTS = 0x0010;
    static const int OPCODE_READ   = 0x00;
    static const int OPCODE_WRITE  = 0x01;

    CHWMemory();
    const char *GetName() const;
    bool InterruptDispatcher(int IntNo);
    bool Step();
};

