#include <cstdlib>
#include <cstdio>
#include <cstring>
#include "CHWMyCPU.h"
#include "CHWMemory.h"
#include "CLog.h"

CHWMyCPU::CHWMyCPU()
{
  // Just init the CPU class
  this->InitCPU();
}

bool
CHWMyCPU::Exec_Init()
{
  // Clear the ROM
  memset(this->ROM, 0, sizeof(this->ROM));

  // Clear the registers
  memset(this->Reg, 0, sizeof(this->Reg));

  // Clear the flags
  this->CPUHalt = false;

  // Debug
  CLOG(LOG_VERBOSE, "Acquiring port %.4x", PORT_RECV_RAM);

  // Register a port for receiving data
  CS::AcquirePort(PORT_RECV_RAM, this);
  CS::System->Ports[PORT_RECV_RAM].SetType(CSPort::PORT_BUFFERED);
  CS::System->Ports[PORT_RECV_RAM].SetBufferSize(0x10);

  // Done
  return true;
}

bool
CHWMyCPU::Exec_Decode()
{
  // Halt?
  if(this->CPUHalt)
    return true; // All OK

  // Is the value sane?
  uint16_t pc = this->Reg[REG_PC];
  if(pc >= sizeof(this->ROM))
  {
    // Throw some exception
    return false;
  }

  // What kinda opcode do we have?
  uint8_t Opcode = this->ROM[pc];
  switch(Opcode)
  {
    case OP_MOV:
      this->CurrentOpcode = OP_MOV;
      this->CurrentOpcodeArg[0] = ARG_REG;
      this->CurrentOpcodeArg[1] = ARG_REG;      
      this->CurrentOpcodeArgValue[0].Reg = (enum Register)this->ROM[pc + 1]; // XXX
      this->CurrentOpcodeArgValue[1].Reg = (enum Register)this->ROM[pc + 2]; // fixme
      pc += 3;
      break;

    case OP_ADD:
      this->CurrentOpcode = OP_ADD;
      this->CurrentOpcodeArg[0] = ARG_REG;
      this->CurrentOpcodeArg[1] = ARG_REG;      
      this->CurrentOpcodeArgValue[0].Reg = (enum Register)this->ROM[pc + 1]; // XXX
      this->CurrentOpcodeArgValue[1].Reg = (enum Register)this->ROM[pc + 2]; // fixme
      pc += 3;
      break;

    case OP_LDI:
      this->CurrentOpcode = OP_LDI;
      this->CurrentOpcodeArg[0] = ARG_REG;
      this->CurrentOpcodeArg[1] = ARG_IMM16; 
      this->CurrentOpcodeArgValue[0].Reg   = (enum Register)this->ROM[pc + 1]; // XXX
      this->CurrentOpcodeArgValue[1].Imm16 = *(uint16_t*)&this->ROM[pc + 2];
      pc += 4;
      break;

    case OP_LDR:
      this->CurrentOpcode = OP_LDR;
      this->CurrentOpcodeArg[0] = ARG_REG;
      this->CurrentOpcodeArg[1] = ARG_ADDR; 
      this->CurrentOpcodeArgValue[0].Reg  = (enum Register)this->ROM[pc + 1]; // XXX
      this->CurrentOpcodeArgValue[1].Addr = *(uint16_t*)&this->ROM[pc + 2];
      pc += 4;
      break;      
 
    case OP_HLT:
      this->CurrentOpcode = OP_HLT;
      break;

    default:
      return false;
  }

  // Save PC
  this->Reg[REG_PC] = pc;

  // Done
  return true;
}

bool
CHWMyCPU::Exec_Execute()
{
  // Halt?
  if(this->CPUHalt)
    return true; // All OK

  // Which instruction is this?
  switch(this->CurrentOpcode)
  {
    case OP_MOV:
      this->Reg[this->CurrentOpcodeArgValue[0].Reg] = 
        this->Reg[this->CurrentOpcodeArgValue[1].Reg];
      break;

    case OP_ADD:
      this->Reg[this->CurrentOpcodeArgValue[0].Reg] += 
        this->Reg[this->CurrentOpcodeArgValue[1].Reg];
      break;

    case OP_LDI:
      this->Reg[this->CurrentOpcodeArgValue[0].Reg] = 
        this->CurrentOpcodeArgValue[1].Imm16;
      break;

    case OP_LDR:
      this->RequestRAMData(
          &this->Reg[this->CurrentOpcodeArgValue[0].Reg],
          PORT_RECV_RAM,
          this->CurrentOpcodeArgValue[1].Addr,
          2);
      break;

    case OP_HLT:
      // Halt the execution
      this->CPUHalt = true; 
      break;

    default:
      // Unsupported?
      return false;
  }

  // Debug
  CLOG(LOG_SUPER_VERBOSE, "CPU state:\n"
         "  r0=%.4x    r1=%.4x    r2=%.4x    r3=%.4x\n"
         "  r4=%.4x    r5=%.4x    r6=%.4x    pc=%.4x",
         this->Reg[REG_R0],
         this->Reg[REG_R1],
         this->Reg[REG_R2],
         this->Reg[REG_R3],
         this->Reg[REG_R4],
         this->Reg[REG_R5],
         this->Reg[REG_R6],
         this->Reg[REG_PC]);

  // Done!
  return true;
}

const char*
CHWMyCPU::GetName() const
{
  // Return
  return "HW_MyCPU";
}

bool
CHWMyCPU::LoadROM(const char *FileName)
{
  // Open the file
  FILE *f = fopen(FileName, "rb");
  if(!f)
  {
    // Some logging
    return false;
  }

  // Read ROM
  fread(this->ROM, 1, sizeof(this->ROM), f);
  
  // Go to the end
  fseek(f, 0, SEEK_END);

  // Was the whole ROM read?
  if(ftell(f) > (long)sizeof(this->ROM))
  {
    // Report that there was not enough space
    // in the ROM memory
  }

  // Done!
  fclose(f);
  return true;
}


// Execution-mode-only function!
bool 
CHWMyCPU::RequestRAMData(void *Buffer, int RecvPort, uint16_t Addr, uint16_t Len)
{
  // Max 16-bytes currently supported
  if(Len > 0x10)
    return false;

  // Send the request
  uint8_t CommandBuffer[6] = 
  {
    CHWMemory::OPCODE_READ,        // Read
    (uint8_t)(Addr & 0xff),        // Lo-Addr
    (uint8_t)((Addr >> 8) & 0xff), // Hi-addr
    (uint8_t)Len,                  // Length
    (uint8_t)(RecvPort & 0xff),        // Lo-Port
    (uint8_t)((RecvPort >> 8) & 0xff)  // Hi-Port
  };

  int i;
  for(i = 0; i < 6; i++) 
    CS::System->Ports[CHWMemory::PORT_REQUESTS].Send(CommandBuffer[i]);

  // Cool, now, go to sleep and wait for the data
  uint8_t *ByteBuffer = (uint8_t*)Buffer;
  do
  {
    // Await next cycle
    this->YieldExec(); 
    
    // Try to receive the data
    while(Len != 0)
    {
      // Try to get the next byte
      if(!CS::System->Ports[RecvPort].Recv(ByteBuffer))
        break; // No luck

      // Lucky! Another byte is ours!
      Len--;
      ByteBuffer++;
    }

    // Continue or stop :)
  } while(Len != 0);

  // Guess everything is received! Done!
  return true;
}

